@@ -8,23 +8,7 @@ use WWW::Crawler::Mojo::UserAgent;
 use Mojo::Message::Request;
 use Mojo::Util qw{md5_sum xml_escape dumper};
 use List::Util;
-our $VERSION = '0.09';
-
-our %tag_attributes = (
-    script  => ['src'],
-    link    => ['href'],
-    a       => ['href'],
-    img     => ['src'],
-    area    => ['href', 'ping'],
-    embed   => ['src'],
-    frame   => ['src'],
-    iframe  => ['src'],
-    input   => ['src'],
-    object  => ['data'],
-    form    => [sub {
-        _weave_form_data(@_);
-    }],
-);
+our $VERSION = '0.10';
 
 has active_conn => 0;
 has fix => sub { {} };
@@ -38,7 +22,62 @@ has 'ua' => sub { WWW::Crawler::Mojo::UserAgent->new };
 has 'ua_name' =>
     "www-crawler-mojo/$VERSION (+https://github.com/jamadam/www-crawler-mojo)";
 has 'shuffle';
-has tag_attributes => sub { \%tag_attributes };
+has element_handlers => sub { {
+    'script[src]'   => sub { $_[0]->{src} },
+    'link[href]'    => sub { $_[0]->{href} },
+    'a[href]'       => sub { $_[0]->{href} },
+    'img[src]'      => sub { $_[0]->{src} },
+    'area'          => sub { $_[0]->{href}, $_[0]->{ping} },
+    'embed[src]'    => sub { $_[0]->{src} },
+    'frame[src]'    => sub { $_[0]->{src} },
+    'iframe[src]'   => sub { $_[0]->{src} },
+    'input[src]'    => sub { $_[0]->{src} },
+    'object[data]'  => sub { $_[0]->{data} },
+    'form'          => sub {
+        my $dom = shift;
+        my (%seed, $submit);
+        
+        $dom->find("[name]")->each(sub {
+            my $e = shift;
+            $seed{my $name = $e->{name}} ||= [];
+            
+            if ($e->type eq 'select') {
+                $e->find('option[selected]')->each(sub {
+                    push(@{$seed{$name}}, shift->{value});
+                });
+            } elsif ($e->type eq 'textarea') {
+                push(@{$seed{$name}}, $e->text);
+            }
+            
+            return unless (my $type = $e->{type});
+            
+            if (!$submit && grep{$_ eq $type} qw{submit image}) {
+                $submit = 1;
+                push(@{$seed{$name}}, $e->{value});
+            } elsif (grep {$_ eq $type} qw{text hidden number}) {
+                push(@{$seed{$name}}, $e->{value});
+            } elsif (grep {$_ eq $type} qw{checkbox}) {
+                push(@{$seed{$name}}, $e->{value}) if (exists $e->{checked});
+            } elsif (grep {$_ eq $type} qw{radio}) {
+                push(@{$seed{$name}}, $e->{value}) if (exists $e->{checked});
+            }
+        });
+        
+        return [$dom->{action},
+                    uc ($dom->{method} || 'GET'), Mojo::Parameters->new(%seed)];
+    },
+    'meta[content]' => sub {
+        return $1 if ($_[0] =~ qr{http\-equiv="?Refresh"?}i &&
+                                (($_[0]->{content} || '') =~ qr{URL=(.+)}i)[0]);
+    },
+    'style' => sub {
+        my $dom = shift;
+        return collect_urls_css($dom->content);
+    },
+    '[style]' => sub {
+        collect_urls_css(shift->{style});
+    }
+} };
 
 sub crawl {
     my ($self) = @_;
@@ -172,46 +211,29 @@ sub scrape {
     my $base = $job->resolved_uri;
     my $type = $res->headers->content_type;
     
-    if ($type && $type =~ qr{text/(html|xml)} &&
-                                (my $base_tag = $res->dom->at('base'))) {
-        $base = resolve_href($base, $base_tag->attr('href'));
-    }
-    
-    my $cb = sub {
-        my ($url, $dom, $method, $params) = @_;
-        
-        $url = _clean_url_obj($url);
-        my $resolved = resolve_href($base, $url);
-        
-        return unless ($resolved->scheme =~ qr{http|https|ftp|ws|wss});
-        
-        my $child = $job->child(resolved_uri => $resolved, literal_uri => $url);
-        
-        $child->method($method) if $method;
-        
-        if ($params) {
-            if ($method eq 'GET') {
-                $child->resolved_uri->query->append($params);
-            } else {
-                $child->tx_params($params);
-            }
-        }
-        
-        $self->emit('refer', sub {
-            $self->enqueue($_[0] || $child);
-        }, $child, $dom || $job->resolved_uri);
-    };
-    
     if ($type && $type =~ qr{text/(html|xml)}) {
+        if ((my $base_tag = $res->dom->at('base'))) {
+            $base = resolve_href($base, $base_tag->attr('href'));
+        }
         my $encode = guess_encoding($res) || 'utf-8';
-        my $body = Encode::decode($encode, $res->body);
-        $self->collect_urls_html(Mojo::DOM->new($body), $cb);
+        my $dom = Mojo::DOM->new(Encode::decode($encode, $res->body));
+        for my $selector (sort keys %{$self->element_handlers}) {
+            $dom->find($selector)->each(sub {
+                my $dom = shift;
+                return if ($dom->xml && _wrong_dom_detection($dom));
+                for ($self->element_handlers->{$selector}->($dom)) {
+                    $self->_delegate_enqueue($_, $dom, $job, $base);
+                }
+            });
+        }
     }
     
     if ($type && $type =~ qr{text/css}) {
         my $encode  = guess_encoding($res) || 'utf-8';
         my $body    = Encode::decode($encode, $res->body);
-        $self->collect_urls_css($body, $cb);
+        for (collect_urls_css($body)) {
+            $self->_delegate_enqueue($_, $job->resolved_uri, $job, $base);
+        }
     }
 };
 
@@ -242,40 +264,35 @@ sub _enqueue {
     }
 }
 
-sub collect_urls_html {
-    my ($self, $dom, $cb) = @_;
+sub _delegate_enqueue {
+    my ($self, $url, $dom, $job, $base) = @_;
+    my $method, my $params;
     
-    $dom->find(join(',', keys %{$self->tag_attributes}))->each(sub {
-        my $dom = shift;
-        return if ($dom->xml && _wrong_dom_detection($dom));
-        for (@{$self->tag_attributes->{$dom->type}}) {
-            if (ref $_ && ref $_ eq 'CODE') {
-                $cb->($_->($dom));
-            }
-            $cb->($dom->{$_}, $dom) if ($dom->{$_});
+    return unless $url;
+    ($url, $method, $params) = @$url if (ref $url);
+    
+    $url = _clean_url_obj($url);
+    my $resolved = resolve_href($base, $url);
+    
+    return unless ($resolved->scheme =~ qr{http|https|ftp|ws|wss});
+    
+    my $child = $job->child(resolved_uri => $resolved, literal_uri => $url);
+    
+    $child->method($method) if $method;
+    
+    if ($params) {
+        if ($method eq 'GET') {
+            $child->resolved_uri->query->append($params);
+        } else {
+            $child->tx_params($params);
         }
-    });
-    $dom->find('meta[http\-equiv=Refresh]')->each(sub {
-        my $dom = shift;
-        if (my $href = $dom->{content} && ($dom->{content} =~ qr{URL=(.+)}i)[0]) {
-            $cb->($href, $dom);
-        }
-    });
-    $dom->find('style')->each(sub {
-        my $dom = shift;
-        $self->collect_urls_css($dom->content || '', sub { $cb->(shift, $dom) });
-    });
-    $dom->find('*[style]')->each(sub {
-        my $dom = shift;
-        $self->collect_urls_css($dom->{style}, sub { $cb->(shift, $dom) });
-    });
+    }
+    
+    $self->emit('refer', sub { $self->enqueue($_[0] || $child) }, $child, $dom);
 }
 
 sub collect_urls_css {
-    my ($self, $str, $cb) = @_;
-    $str =~ s{/\*.+?\*/}{}gs;
-    my @urls = ($str =~ m{url\(['"]?(.+?)['"]?\)}g);
-    $cb->($_) for (@urls);
+    map { s/^(['"])// && s/$1$//; $_ } (shift || '') =~ m{url\((.+?)\)}ig;
 }
 
 my $charset_re = qr{\bcharset\s*=\s*['"]?([a-zA-Z0-9_\-]+)['"]?}i;
@@ -302,43 +319,6 @@ sub resolve_href {
     return $abs;
 }
 
-sub _weave_form_data {
-    my $form = shift;
-    my %seed;
-    my $submit;
-    
-    $form->find("[name]")->each(sub {
-        my $e = shift;
-        my $name = $e->attr('name');
-        $seed{$name} ||= [];
-        
-        if ($e->type eq 'select') {
-            $e->find('option[selected=selected]')->each(sub {
-                push(@{$seed{$name}}, shift->attr('value'));
-            });
-        } elsif ($e->type eq 'textarea') {
-            push(@{$seed{$name}}, $e->text);
-        }
-        
-        my $type = $e->attr('type');
-        return unless $type;
-        
-        if (!$submit && grep{$_ eq $type} qw{submit image}) {
-            $submit = 1;
-            push(@{$seed{$name}}, $e->attr('value'));
-        } elsif (grep {$_ eq $type} qw{text hidden number}) {
-            push(@{$seed{$name}}, $e->attr('value'));
-        } elsif (grep {$_ eq $type} qw{checkbox}) {
-            push(@{$seed{$name}}, $e->attr('value')) if (exists $e->{checked});
-        } elsif (grep {$_ eq $type} qw{radio}) {
-            push(@{$seed{$name}}, $e->attr('value')) if (exists $e->{checked});
-        }
-    });
-    
-    return $form->{action}, $form,
-                    uc ($form->{method} || 'GET'), Mojo::Parameters->new(%seed);
-}
-
 sub _urls_redirect {
     my $tx = shift;
     my @urls;
@@ -391,16 +371,10 @@ sub _mod_busyness {
 
 sub _host_key {
     state $well_known_ports = {http => 80, https => 443};
-    
-    my $uri = ref $_[0] ? $_[0] : Mojo::URL->new($_[0]);
+    my $uri = shift;
     my $key = $uri->scheme. '://'. $uri->ihost;
-    
-    if (my $port = $uri->port) {
-        if ($port ne $well_known_ports->{$uri->scheme}) {
-            $key .= ':'. $port;
-        }
-    }
-    
+    return $key unless (my $port = $uri->port);
+    $key .= ':'. $port if ($port ne $well_known_ports->{$uri->scheme});
     return $key;
 }
 
@@ -446,10 +420,15 @@ moderate range of web pages so DO NOT use it for persistent crawler jobs.
 L<WWW::Crawler::Mojo> inherits all attributes from L<Mojo::EventEmitter> and
 implements the following new ones.
 
-=head2 tag_attributes
+=head2 element_handlers
 
-Catalog of HTML attribute names which possibly contain URLs. Defaults to
-%__PACKAGE__::tag_attributes.
+HTML element handler on scraping.
+
+    my $handlers = $bot->element_handlers;
+    $bot->element_handlers->{img} = sub {
+        my $dom = shift;
+        return $dom->{src};
+    };
 
 =head2 ua
 
@@ -534,7 +513,7 @@ implements the following new ones.
 
 =head2 res
 
-Emitted when crawler got response from server.
+Emitted when crawler got response from server. The callback takes 4 arguments.
 
     $bot->on(res => sub {
         my ($bot, $scrape, $job, $res) = @_;
@@ -545,6 +524,26 @@ Emitted when crawler got response from server.
         }
     });
 
+=over
+
+=item $bot
+
+L<WWW::Crawler::Mojo> instance.
+
+=item $scrape
+
+Scrape URLs out of the document. this is a shorthand of $bot->scrape($job)
+
+=item $job
+
+L<WWW::Crawler::Mojo::Job> instance.
+
+=item $res
+
+L<Mojo::Message::Response> instance.
+
+=back
+
 =head2 refer
 
 Emitted when new URI is found. You can enqueue the URI conditionally with
@@ -561,6 +560,26 @@ the callback.
         }
     });
 
+=over
+
+=item $bot
+
+L<WWW::Crawler::Mojo> instance.
+
+=item $enqueue
+
+Scrape URLs out of the document. this is a shorthand of $bot->enqueue($job)
+
+=item $job
+
+L<WWW::Crawler::Mojo::Job> instance.
+
+=item $context
+
+Either L<Mojo::DOM> or L<Mojo::URL> instance.
+
+=back
+
 =head2 empty
 
 Emitted when queue length got zero. The length is checked every 5 seconds.
@@ -678,9 +697,7 @@ Collects URLs out of HTML.
 
 Collects URLs out of CSS.
 
-    $bot->collect_urls_css($dom, sub {
-        my $uri = shift;
-    });
+    @urls = collect_urls_css($dom);
 
 =head2 guess_encoding
 
@@ -694,24 +711,6 @@ Resolves URLs with a base URL.
 
     WWW::Crawler::Mojo::resolve_href($base, $uri);
 
-=head1 CONSTANTS
-
-=head2 %tag_attributes
-
-A catalog of HTML attribute names which possibly contain URLs.
-
-    script  => ['src'],
-    link    => ['href'],
-    a       => ['href'],
-    img     => ['src'],
-    area    => ['href', 'ping'],
-    embed   => ['src'],
-    frame   => ['src'],
-    iframe  => ['src'],
-    input   => ['src'],
-    object  => ['data'],
-    form    => ['action'],
-
 =head1 EXAMPLE
 
 L<https://github.com/jamadam/WWW-Flatten>
